
import enum


@enum.unique
class ConnectionStatus(enum.Enum):
    NOT_APPLICABLE = "NOT_APPLICABLE"
    CALCULATING = "CALCULATING"
    CONNECTED = "CONNECTED"
    DISCONNECTED = "DISCONNECTED"
    DOMAIN_DELETED = "DOMAIN_DELETED"
    NOT_COMPLETE = "NOT_COMPLETE"
    USER_IN_SOURCE = "USER_IN_SOURCE"
    L7_APPLICATION = "L7_APPLICATION"
    ANY_OR_A_CLASS = "ANY_OR_A_CLASS"
    INTERNET = "INTERNET"
    NON_CONTINUOUS_MASK = "NON_CONTINUOUS_MASK"
